/* 
  mesh.c

  Contains routines for computing mesh estimates of the value and error.

*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "functions.h"
#include "global.h"
#include "util.h"
#include "plist.h"
#include "blackScholes.h"

/* Prototypes for mesh estimator */
double cluster_mesh_estimate0(POINT *parent, POINT *point, double *error);
double cluster_mesh_estimate1(POINT *parent, POINT *point, double *error);
double cluster_mesh_estimate3(POINT *parent, POINT *point, double *error);

/*  - - - - - - - - - -
   Routine to compute mesh estimate.
   Uses Time and which_inner_control from global.h
*/
double mesh_estimate_value(POINT *point, double *error){
  POINT *parent;
  int i, parent_id; 
  double sigma, I, sum_sigma, sum_I, sum_weight, w;
  sum_sigma = 0.0;
  sum_I = 0.0;
  sum_weight = 0.0;

  for(i=0; i<point->important_cluster.elements; i++){
    parent_id =  point->important_cluster.element[i];
    parent = &(Time[point->timeslice].element[parent_id]);

    if (parent->children.elements) { 
 
      switch(which_inner_control){

      case(1):
	I = cluster_mesh_estimate1(parent, point, error);
    	break;
      case(3):
	I = cluster_mesh_estimate3(parent, point, error);
    	break;
      default:
	I = cluster_mesh_estimate0(parent, point, error);
      } /*end switch */
	
      if(*error == 0.0 && I != 0.0){
	/*       	  printf("WARNING: mesh_estimate_value has zero error\n"); */
      	return(I);
      }

      if(fabs(*error) > SMALLVAL && fabs(I) > SMALLVAL){
	w = 1/((*error) * (*error));
	sum_weight += w;
	sum_I += I * w;
	sum_sigma += w*w*(*error)*(*error);
	  /*++sum_sigma;*/
      }
    }
  }
  if(fabs(sum_weight) < SMALLVAL && fabs(sum_I) > SMALLVAL){
    *error = 0;
    return(sum_I);
  }

  if(fabs(sum_I) < SMALLVAL && fabs(sum_weight) < SMALLVAL ){
    *error = 0;
    return(0);
  }
  *error = sqrt(sum_sigma)/sum_weight;
  return(sum_I/sum_weight);
	
} /* end mesh_estimate_value */

/* 
 * cluster_mesh_estimate0
 * Routine to compute mesh estimates without controls for clusters.
 */

double cluster_mesh_estimate0(POINT *parent, POINT *point, double *error){
  int i;
  double *from, *to, *center;
  double f,w,S1,S2,S3,S4,S5,S6;

  S1 = 0;
  S2 = 0;
  S3 = 0;
  S4 = 0;
  S5 = 0;
  S6 = 0;

  from = point->price;
  center = parent->price;

  /* Loop over children in parent (cluster) */  
  for(i=0;i<parent->children.elements; i++){
    /* get the probability of going from point to point i in cluster */
    to = Time[parent->timeslice+1].element[parent->children.element[i]].price;

    w = transition_probability(from, to) / 
      transition_probability(center,to);

/*     printf("CMS: child = %d weight = %lf\n", i, w); */

    /* value */
    f = Time[parent->timeslice+1].element[parent->children.element[i]].value;
    
    S1 += w;
    S2 += w*w;
    S3 += w*f;
    S4 += w*w*f;
    S6 += w*w*f*f;
  }
  *error = ((1/(S1*S1))*(S6 - 2*S3/S1*S4 + (S3/S1)*(S3/S1)*S2));
  if(*error < 0.0){
    *error = 0.0;
  }else{
    *error = sqrt(*error);
  }
  
  /*   printf("cluster_mesh_estimate: value = %lf error = %lf\n", S3/S1, *error); */
  return(S3/S1);    
} /* end cluster_mesh_estimate0 */


/* 
 * cluster_mesh_estimate0
 * Routine to compute mesh estimates without controls for clusters.
 */

double cluster_mesh_estimate1(POINT *parent, POINT *point, double *error){
  double *from, *to, *center;
  int max1, i;
  double S1, S2, S7, S8, S9 ,S10, S11, S12, S13, w, f, g, G,
    maxPrice, beta, S14, S15, S16, S17;
  POINT *this;

  S1 = 0;
  S2 = 0;
  S7 = 0;
  S8 = 0;
  S9 = 0;
  S10 = 0;
  S11 = 0;
  S12 = 0;
  S13 = 0;
  S14 = 0;
  S15 = 0;
  S16 = 0;
  S17 = 0.0;

  /* At the given time find the max prices */

  max1 = dWhichMax(Dimensions, point->price);

  G = exp(-point->timeslice*Interest_rate)*
    euroCall(point->price[max1], Strike_price, Interest_rate,
	     sqrt(Covariance[max1][max1]),1.0, Dividend_rate[max1]);

  from = point->price;
  center = parent->price;

  /* Loop over children in parent (cluster) */  
  for(i=0;i<parent->children.elements; i++){
    /* get the probability of going from point to point i in cluster */
    this = &(Time[parent->timeslice+1].element[parent->children.element[i]]);
    to = this->price;

    w = transition_probability(from, to) / 
      transition_probability(center,to);

    /* value */
    f = this->value;
    maxPrice = this->price[max1];
    
    if(maxPrice > Strike_price){
      g = exp(-this->timeslice*Interest_rate)*(maxPrice - Strike_price);
    }else{
      g= 0;
    }

    S1 += w;
    S2 += w*w;
    S7 += w*f;
    S8 += w*(g-G);
    S9 += w*w*f;
    S10 += w*w*(g-G);
    S11 += w*w*f*f;
    S12 += w*w*f*(g-G);
    S13 += w*w*(g-G)*(g-G);
    S14 += w*f*(g-G);
    S15 += w*(g-G)*(g-G);
    S16 += w*f*f;
    S17 += w * g;
  }

  /* Old derivation */
/*   beta = ((S14/S1) - (S7/S1)*(S8/S1))/   */
/*     ((S15/S1) - (S8/S1)*(S8/S1));   */


/*   *error = sqrt(S16/S1 - (S7/S1)*(S7/S1)      */
/* 		-2*beta*(S14/S1 - (S7/S1)*(S8/S1))      */
/* 		+beta*beta*((S15/S1) - (S8/S1)*(S8/S1)));     */


  /* If beta blows up, return uncontrolled value and error */
/*   if((S15/S1) - (S8/S1)*(S8/S1) == 0){ */
/*     *error = sqrt((1/(S1*S1))*(S11 - 2*S7/S1*S9 + (S7/S1)*(S7/S1)*S2)); */
/*     return(S7/S1); */

/*   } */

  /* New derivation */
  if(fabs(S13/S1 - 2*S8*S10/(S1*S1) + S2*S8*S8/(S1*S1*S1)) < SMALLVAL){
       *error = ((1/(S1*S1))*(S11 - 2*S7/S1*S9 + (S7/S1)*(S7/S1)*S2));
       *error = sqrt(fabs(*error));
       return(S7/S1);
     }

     beta = (S12/S1 - (S8*S9 + S7*S10)/(S1*S1) + S2*S7*S8/(S1*S1*S1))/
       (S13/S1 - 2*S8*S10/(S1*S1) + S2*S8*S8/(S1*S1*S1));



     *error = 1/(S1*S1)*((S11 -2*beta*S12 + beta*beta*S13)  
			      -(2/S1)*(S7-beta*S8)*(S9-beta*S10) 
                              +(S2/(S1*S1))*(S7*S7 -2*beta*S7*S8 +beta*beta*S8*S8) );
     if(*error < 0){
       *error = 0.0;
     }else{
       *error = sqrt(*error);
     }  

     return( (S7-beta*S8)/S1);
} /* end cluster_mesh_estimate1 */




double cluster_mesh_estimate3(POINT *parent, POINT *point, double *error){
  double *from, *to, *center;
  int max1, max2, i;
  double S1, S2, S7, S8, S9 ,S10, S11, S12, S13, w, f, g, G,
    maxval1, maxval2, beta, S14, S15, S16, maxPrice;
  POINT *this;

  S1 = 0;
  S2 = 0;
  S7 = 0;
  S8 = 0;
  S9 = 0;
  S10 = 0;
  S11 = 0;
  S12 = 0;
  S13 = 0;
  S14 = 0;
  S15 = 0;
  S16 = 0;

  /* At the given time find the max prices */

  maxval1 = 0.0;
  maxval2 = 0.0;
  max1 = dWhichMax(Dimensions, point->price);
  maxval1 = point->price[max1];

  for(i=0; i<Dimensions;i++){
    if(i != max1 && point->price[i] > maxval2){
      max2 = i;
      maxval2 = point->price[i];
    }
  }


  G = exp(-point->timeslice*Interest_rate)*    
    maxOf2Risk(maxval1, maxval2, Strike_price, 1.0,  
	       sqrt(Covariance[max1][max1]),   
	       sqrt(Covariance[max2][max2]),   
	       Interest_rate,
	       Dividend_rate[max1],   
	       Dividend_rate[max2],  
	       Covariance[max1][max2]); 

  from = point->price;
  center = parent->price;

  /* Loop over children in parent (cluster) */  
  for(i=0;i<parent->children.elements; i++){
    /* get the probability of going from point to point i in cluster */
    this = &(Time[parent->timeslice+1].element[parent->children.element[i]]);
    to = this->price;

    w = transition_probability(from, to) / 
      transition_probability(center,to);

    /* value */
    f = this->value;
        
    if( this->price[max1] >= this->price[max2]){
      maxPrice = this->price[max1];
    }else{
      maxPrice = this->price[max2];
    }
    if(maxPrice > Strike_price){
      g = exp(-this->timeslice*Interest_rate)*(maxPrice - Strike_price);
    }else{
      g= 0;
    }

    S1 += w;
    S2 += w*w;
    S7 += w*f;
    S8 += w*(g-G);
    S9 += w*w*f;
    S10 += w*w*(g-G);
    S11 += w*w*f*f;
    S12 += w*w*f*(g-G);
    S13 += w*w*(g-G)*(g-G);
    S14 += w*f*(g-G);
    S15 += w*(g-G)*(g-G);
    S16 += w*f*f;
  }

  /* Old derivation */
/*   beta = ((S14/S1) - (S7/S1)*(S8/S1))/   */
/*     ((S15/S1) - (S8/S1)*(S8/S1));   */


/*   *error = sqrt(S16/S1 - (S7/S1)*(S7/S1)      */
/* 		-2*beta*(S14/S1 - (S7/S1)*(S8/S1))      */
/* 		+beta*beta*((S15/S1) - (S8/S1)*(S8/S1)));     */


  /* If beta blows up, return uncontrolled value and error */
/*   if((S15/S1) - (S8/S1)*(S8/S1) == 0){ */
/*     *error = sqrt((1/(S1*S1))*(S11 - 2*S7/S1*S9 + (S7/S1)*(S7/S1)*S2)); */
/*     return(S7/S1); */

/*   } */

  /* New derivation */
     beta = (S12/S1 - (S8*S9 + S7*S10)/(S1*S1) + S2*S7*S8/(S1*S1*S1))/ 
       (S13/S1 - 2*S8*S10/(S1*S1) + S2*S8*S8/(S1*S1*S1)); 

     *error = 1/(S1*S1)*((S11 -2*beta*S12 + beta*beta*S13)  
              -(2/S1)*(S7-beta*S8)*(S9-beta*S10)  
              +(S2/(S1*S1))*(S7*S7 -2*beta*S7*S8 +beta*beta*S8*S8) );  
     if(*error < 0){
       *error = 0.0;
     }else{
       *error = sqrt(*error);
     }  

  return( (S7-beta*S8)/S1);
} /* end cluster_mesh_estimate3 */

